using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Example2
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // physics variables

        // time
        float time = 0.0f;

        // mass
        float mass = 1.0f;

        // moment of inertia (scalar for 2D)
        float momentOfInertia = 200.0f;

        // position
        Vector2 position = new Vector2(100.0f, 100.0f);

        // orientation (can use scalar for 2D)
        float orientation = 0.0f;

        // linear velocity
        Vector2 linearVelocity = Vector2.Zero;

        // angular velocity (scalar for 2D)
        float angularVelocity = 0.0f;
     
        // total force
        Vector2 totalForce = Vector2.Zero;

        // total torque (scalar for 2D)
        float totalTorque = 0.0f;

        // user input variables
        bool leftJet;
        bool rightJet;
        bool frictionEnabled;
        const float linearFrictionConstant = 0.5f;
        const float angularFrictionConstant = 300.0f;

        // visualisation variables
        SpriteFont font;
        Vector2 shipCentre;
        Vector2 jetCentre;
        float jetScale = 1.0f;
        Texture2D backgroundTexture;
        Texture2D shipTexture;
        Texture2D jetTexture;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            InitialisePhysicsState();

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // TODO: use this.Content to load your game content here
            font = Content.Load<SpriteFont>("Font");

            backgroundTexture = Content.Load<Texture2D>("Background");
            shipTexture = Content.Load<Texture2D>("Ship");
            jetTexture = Content.Load<Texture2D>("Jet");
            shipCentre = new Vector2(shipTexture.Width * 0.5f, shipTexture.Height * 0.5f);
            jetCentre = new Vector2(jetTexture.Width * 1.0f, jetTexture.Height * 0.5f);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here
            ProcessPhysicsInput();
            UpdatePhysicsState((float) gameTime.ElapsedGameTime.TotalSeconds);

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // visualise simulation
            VisualisePhysicsSimulation();

            base.Draw(gameTime);
        }

        // initialise physics simulation
        private void InitialisePhysicsState()
        {
            // set initial time
            time = 0.0f;

            // set mass (m) and moment of inertia (I) of ship
            mass = 1.0f;
            momentOfInertia = 200.0f;

            // centre ship in middle of screen and set zero orientation
            position = new Vector2(
                graphics.GraphicsDevice.PresentationParameters.BackBufferWidth * 0.5f,
                graphics.GraphicsDevice.PresentationParameters.BackBufferHeight * 0.5f);
            orientation = 0.0f;

            // initialise linear and angular velocities
            linearVelocity = Vector2.Zero;
            angularVelocity = 0.0f;

            // initialise applied force and torque
            totalForce = Vector2.Zero;
            totalTorque = 1.0f;
        }

        // processes the user input and generates appropriate forces
        // on the ship
        private void ProcessPhysicsInput()
        {
            // compute orientation matrix
            Matrix orientationMatrix = Matrix.CreateRotationZ(orientation);

            totalForce = Vector2.Zero;
            totalTorque = 0.0f;

            KeyboardState keyboardState = Keyboard.GetState();

            if (keyboardState.IsKeyDown(Keys.Escape))
            {
                InitialisePhysicsState();
            }

            // left / forward control
            if (keyboardState.IsKeyDown(Keys.Left) || keyboardState.IsKeyDown(Keys.Up))
            {
                // apply right jet force and torque
                Vector2 rightJetForce = new Vector2(50.0f, 00.0f);
                rightJetForce = Vector2.Transform(rightJetForce, orientationMatrix);
                totalForce += rightJetForce;

                totalTorque -= 50.0f * 21.0f;

                // set flag for drawing later
                rightJet = true;
            }
            else
                rightJet = false;

            // right / forward control
            if (keyboardState.IsKeyDown(Keys.Right) || keyboardState.IsKeyDown(Keys.Up))
            {
                // apply left jet force and torque
                Vector2 leftJetForce = new Vector2(50.0f, 00.0f);
                leftJetForce = Vector2.Transform(leftJetForce, orientationMatrix);
                totalForce += leftJetForce;

                totalTorque += 50.0f * 21.0f;

                // set flag for drawing later
                leftJet = true;
            }
            else
                leftJet = false;

            // friction controls
            if (keyboardState.IsKeyDown(Keys.F))
                frictionEnabled = true;
            if (keyboardState.IsKeyDown(Keys.N))
                frictionEnabled = false;

            // if friction enabled, simulate simple 'aerodynamic' friction
            // by adding forces opposing velocity
            // (the size if this force is proportional to velocity)
            if (frictionEnabled)
            {
                // add resistive force
                totalForce -= linearVelocity * linearFrictionConstant;

                // add resisitive torque
                totalTorque -= angularVelocity * angularFrictionConstant;
            }
        }

        // updates the physical state of the simulation
        private void UpdatePhysicsState(float deltaT)
        {
            // update time
            time += deltaT;

            // compute linear acceleration (a = f / m)
            Vector2 linearAcceleration = totalForce / mass;

            // compute angular acceleration (alpha = tau / I)
            float angularAcceleration = totalTorque / momentOfInertia;

            // compute new linear velocity (v' = v + a * Dt)
            Vector2 newLinearVelocity = linearVelocity + linearAcceleration * deltaT;

            // compute new angular velocity (omega' = omega + alpha * Dt)
            float newAngularVelocity = angularVelocity + angularAcceleration * deltaT;

            // compute new position (p' = p + v * Dt)
            Vector2 newPosition = position + linearVelocity * deltaT;

            // compute new orientation (phi' = phi + omega * Dt)
            float newOrientation = orientation + angularVelocity * deltaT;

            // NOTE: must restrict orientation between -PI and +PI
            while (newOrientation > MathHelper.Pi)
                newOrientation -= MathHelper.TwoPi;
            while (newOrientation < -MathHelper.Pi)
                newOrientation += MathHelper.TwoPi;

            // assign new linear and angular positions and velocities
            position = newPosition;
            orientation = newOrientation;
            linearVelocity = newLinearVelocity;
            angularVelocity = newAngularVelocity;

            // do **very simple** collision detection / response with window borders
            if (position.X < 0 || position.X >= graphics.GraphicsDevice.PresentationParameters.BackBufferWidth)
                linearVelocity.X = -linearVelocity.X;
            if (position.Y < 0 || position.Y >= graphics.GraphicsDevice.PresentationParameters.BackBufferHeight)
                linearVelocity.Y = -linearVelocity.Y;
        }

        // visualise simulation on screen
        private void VisualisePhysicsSimulation()
        {
            // compute orientation matrix for drawing
            Matrix orientationMatrix = Matrix.CreateRotationZ(orientation);

            // compute an oscillating scaling factor to animate the ship jets
            // (this is just eye candy to make the jets look nice)
            jetScale = 0.9f + (float)Math.Cos(time * MathHelper.Pi * 8.0f) * 0.1f;

            // start rendering
            spriteBatch.Begin();

            // draw starfield backgroud
            spriteBatch.Draw(backgroundTexture, Vector2.Zero, Color.White);

            // draw ship with current position and orientation
            spriteBatch.Draw(shipTexture, position, null, Color.White, orientation, shipCentre, 1.0f, SpriteEffects.None, 0.0f);

            // if left jet active
            if (leftJet)
            {
                // determine jet position (without considering orientation)
                Vector2 jetPosition = new Vector2(-24, -21);

                // rotate offset according to orientation
                jetPosition = Vector2.Transform(jetPosition, orientationMatrix);

                // translate offset to centre around ship
                jetPosition += position;

                // draw jet sprite
                spriteBatch.Draw(jetTexture, jetPosition, null, Color.White, orientation, jetCentre, jetScale, SpriteEffects.None, 0.0f);
            }

            if (rightJet)
            {
                // determine jet position (without considering orientation)
                Vector2 jetPosition = new Vector2(-24, 21);

                // rotate offset according to orientation
                jetPosition = Vector2.Transform(jetPosition, orientationMatrix);

                // translate offset to centre around ship
                jetPosition += position;

                // draw jet sprite
                spriteBatch.Draw(jetTexture, jetPosition, null, Color.White, orientation, jetCentre, jetScale, SpriteEffects.None, 0.0f);
            }

            // end of rendering
            spriteBatch.End();

            // HUD info display
            DrawText(new Vector2(16, 16), "PHYSICS-BASED ANIMATION - EXAMPLE 2");
            DrawText(new Vector2(16, 48), "Press ESC to reset simulation");
            DrawText(new Vector2(16, 64), "Press UP to switch both jets");
            DrawText(new Vector2(16, 80), "Press LEFT to switch right jet");
            DrawText(new Vector2(16, 96), "Press RIGHT to switch left jet");
            DrawText(new Vector2(16, 112), frictionEnabled ? "Press N to disable friction" : "Press F to enable friction");
        }

        // private method for drawing white text with shadow
        private void DrawText(Vector2 textPosition, string textMessage)
        {
            spriteBatch.Begin();
            Vector2 shadowPosition = textPosition + new Vector2(1.0f, 1.0f);
            spriteBatch.DrawString(font, textMessage, shadowPosition, Color.Red);
            spriteBatch.DrawString(font, textMessage, textPosition, Color.White);
            spriteBatch.End();
        }
    }
}