﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Example3
{
    // represents a body in the physicsl simulation (implemented as a circle)
    class Body
    {
        // density constant used to compute mass from circle area
        // can be varied to make bodies lighter or heavier
        private const float DensityFactor = 0.01f;

        // position and velocity vectors
        private Vector m_position;
        private Vector m_velocity;

        // body size (circle radius)
        private float m_radius;

        // force currently applied on body
        private Vector m_force;

        // constructs a body with the given position, velocity and radius
        public Body(Vector position, Vector velocity, float radius)
        {
            m_position = position;
            m_velocity = velocity;
            m_radius = radius;
        }

        // returns true of the body contains the given point
        public bool ContainsPoint(Vector point)
        {
            // sqaured comparison avoids costly square root
            return (point - m_position).LengthSquared <= m_radius * m_radius;
        }

        // updates the position and velocity over the given delta time
        public void Update(float deltaTime)
        {
            // determine the acceleration from the currently
            // applied force
            Vector acceleration = m_force / Mass;

            // compute new position and velocity
            Vector newPosition = m_position + m_velocity * deltaTime;
            Vector newVelocity = m_velocity + acceleration * deltaTime;

            // replace old position and velocity with new values
            m_position = newPosition;
            m_velocity = newVelocity;
        }

        // position property
        public Vector Position
        {
            get { return m_position; }
            set { m_position = value; }
        }

        // velocity property
        public Vector Velocity
        {
            get { return m_velocity; }
            set { m_velocity = value; }
        }

        // radius property
        public float Radius
        {
            get { return m_radius; }
            set { m_radius = value; }
        }

        // applied force property
        public Vector Force
        {
            get { return m_force; }
            set { m_force = value; }
        }

        // squared radius property
        public double RadiusSquared
        {
            get { return (double)m_radius * m_radius; }
        }

        // mass property
        public float Mass
        {
            // mass computed as Area * Density Factor
            get { return DensityFactor * (float) Math.PI * m_radius * m_radius; }
        }
    }
}
