﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Example3
{
    // 2D vector implementation with overloaded operators
    // for addition, subtraction, scalar multiplication,
    // dot product and normalisation operations
    struct Vector
    {
        // vector components
        public float X, Y;

        // negation operator v2 = -v
        public static Vector operator -(Vector vector)
        {
            return new Vector(-vector.X, -vector.Y);
        }

        // plus operator v = v1 + v2
        public static Vector operator +(Vector vector1, Vector vector2)
        {
            return new Vector(vector1.X + vector2.X, vector1.Y + vector2.Y);
        }

        // minus operator v = v1 - v2
        public static Vector operator -(Vector vector1, Vector vector2)
        {
            return new Vector(vector1.X - vector2.X, vector1.Y - vector2.Y);
        }

        // dot product d = v1 . v2
        public static float operator *(Vector vector1, Vector vector2)
        {
            return vector1.X * vector2.X + vector1.Y * vector2.Y;
        }

        // vector-scalar multiplication v2 = v * s
        public static Vector operator *(Vector vector, float scalar)
        {
            return new Vector(vector.X * scalar, vector.Y * scalar);
        }

        // scalar-vector multiplication v2 = s * v
        public static Vector operator *(float scalar, Vector vector)
        {
            return new Vector(vector.X * scalar, vector.Y * scalar);
        }

        // scalar division v2 = v / s
        public static Vector operator /(Vector vector, float scalar)
        {
            return new Vector(vector.X / scalar, vector.Y / scalar);
        }

        // static property for zero vector
        public static Vector Zero { get { return new Vector(); } }

        // constructs a vector with the given components
        public Vector(float x, float y)
        {
            X = x;
            Y = y;
        }

        // vector magnitude
        public float Length
        {
            get
            {
                // work in double for extended range
                return (float)Math.Sqrt((double)X * X + (double)Y * Y); ;
            }
        }

        // vector magnitude squared
        public double LengthSquared
        {
            get
            {
                return (double)X * X + (double)Y * Y;
            }
        }

        // normal form of vector (same direction, but unit length)
        public Vector Normal
        {
            get
            {
                return this / Length;
            }
        }

        // perpendicular version of vector (rotated by 90 degs)
        public Vector Perpendicular
        {
            get
            {
                return new Vector(-Y, X);
            }
        }
    }
}
